import asyncio
import json
import logging
from os import environ as env
from os import path as path
from time import time

import botocore
import pscr_common.config as pscr_config
import requests
from pscr_common.aws import get_boto3_client, get_ssm_parameter
from pscr_common.log import setup_logging, clear_logging_handlers
from pscr_common.metadata import MetadataProvider
from pscr_common.metrics import get_reporter
from pyformance import global_registry
from requests.adapters import HTTPAdapter, Retry

_logger = logging.getLogger(__name__)


# TODO: Remove when lambda mode is no longer supported
def _temp_setup_lambda():
    if not env.get("LAMBDA"):
        return
    current_env = env.get("ENV_ID")
    domains = {
        "dev": "https://defense-dev01.cbdtest.io",
        "staging": "https://defense-test03.cbdtest.io",
        "usprod": "https://defense-prod05.conferdeploy.net",
        "euprod": "https://defense-eu.conferdeploy.net",
        "apjprod": "https://defense-prodnrt.conferdeploy.net",
        "ausprod": "https://defense-prodsyd.conferdeploy.net",
    }
    if current_env in domains.keys():
        env["ENTITLEMENT_SERVICE_AUTH_TOKEN"] = get_ssm_parameter(f"/{current_env}/backend_connector_auth_token")
        env["ENTITLEMENT_SERVICE_BASE_URL"] = domains[current_env]

    env["AWS_S3_BUCKET"] = f"{current_env}-pscr-admin-bucket"
    env["ENTITLEMENT_SERVICE_RESULTS_LIMIT"] = (
        env.get("ENTITLEMENT_SERVICE_RESULTS_LIMIT") if env.get("ENTITLEMENT_SERVICE_RESULTS_LIMIT") else "100"
    )
    env["ENTITLEMENT_SERVICE_FEATURES"] = (
        env.get("ENTITLEMENT_SERVICE_FEATURES")
        if env.get("ENTITLEMENT_SERVICE_FEATURES")
        else "defense,mdforeedr,mdr,threathunter,xdr"
    )


def get_orgs_by_feature_from_entitlement_api(feature, base_url, auth_token, results_limit):
    cursor_route = "/entitlements/v1/orgs?features=psc%3Afeature%3A{}&limit={}".format(feature, results_limit)
    http_headers = {"X-Auth-Token": "{}".format(auth_token), "Content-Type": "application/json"}
    counter = 0
    orgs = []

    while True:
        counter += 1
        orgs_by_features_api_route = base_url + cursor_route
        http_session = requests.Session()
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[500, 502, 503, 504])
        http_session.mount("https://", HTTPAdapter(max_retries=retries))
        response = http_session.get(orgs_by_features_api_route, headers=http_headers, timeout=30)
        response.raise_for_status()
        response = response.json()
        if len(response.get("results", [])) != 0:
            orgs = orgs + [item["org_key"] for item in response["results"]]
            print(
                f"Feature: {feature}\n Pagination Counter: {counter}\n CumulativeOrgCount: {len(orgs)}\n cusorRoute: {base_url + cursor_route}\n"
            )
            cursor_route = response["next"]
        if response["next"] is None:
            break

    return orgs


def push_org_list_file_to_s3(feature, orgs_list, s3_bucket):
    s3_client = get_boto3_client("s3")
    s3_key = f"enabled_orgs/{feature}"
    s3_body = ",".join(orgs_list)
    s3_client.put_object(Body=s3_body, Bucket=s3_bucket, Key=s3_key, ContentType="text/plain")


def refresh_orgs_for_feature(feature, base_url, auth_token, results_limit, s3_bucket, registry):
    orgs_list = get_orgs_by_feature_from_entitlement_api(feature, base_url, auth_token, results_limit)
    push_org_list_file_to_s3(feature, orgs_list, s3_bucket)
    org_feature_reconciliation(feature, orgs_list[:], s3_bucket, "clm_notifications", registry)


def config_value(config, section, variable, set_env=False):
    env_var_key = section + "_" + variable
    if env.get(env_var_key) is not None:
        value = env.get(env_var_key)
    else:
        value = config.get(section, variable, fallback="NO_VALUE_DEFINED")
    if set_env:
        env[env_var_key] = value
    return value


def group_features_by_org(s3_bucket):
    try:
        org_features = {}
        s3_client = get_boto3_client("s3")
        response = s3_client.list_objects_v2(Bucket=s3_bucket, Prefix="enabled_orgs/", StartAfter="enabled_orgs/")
        files = response.get("Contents")
        for file in files:
            feature = file["Key"].split("/")[1]
            orgs = s3_client.get_object(Bucket=s3_bucket, Key=file["Key"])["Body"].read().decode("utf-8")
            if orgs:
                orgs_list = orgs.split(",")
                for org in orgs_list:
                    if org_features.get(org) is None:
                        org_features[org] = [feature]
                    else:
                        org_features[org].append(feature)

        s3_client.put_object(Body=json.dumps(org_features), Bucket=s3_bucket, Key="org_features/features.json")

        _logger.info(f"Successfully Uploaded org feature file")

    except botocore.exceptions.ClientError as error:
        _logger.exception(f"Failed to Upload Org Features Config to S3 due to {error.response['Error']['Code']}")


def org_feature_reconciliation(feature, orgs_list, s3_bucket, s3_key, registry):
    try:
        s3_client = get_boto3_client("s3")
        curr_org_list = (
            s3_client.get_object(Bucket=s3_bucket, Key=f"{s3_key}/enabled_orgs/{feature}")["Body"]
            .read()
            .decode("utf-8")
        )

        if curr_org_list:
            curr_org_list = curr_org_list.split(",")

        for org in curr_org_list[:]:
            if org in orgs_list:
                curr_org_list.remove(org)
                orgs_list.remove(org)

        if orgs_list:
            registry.gauge(f"missing_orgs.count.{feature}").set_value(len(orgs_list))
            _logger.info(f"Total Missing Orgs in {feature} file {len(orgs_list)}")

        if curr_org_list:
            registry.gauge(f"additional_orgs.count.{feature}").set_value(len(curr_org_list))
            _logger.info(f"Total Additional Orgs in {feature} file {len(curr_org_list)}")

    except Exception as exception:
        _logger.warning(f"Exception in org reconciliation function {exception}")


def str_to_bool(value):
    if value.lower() == "true":
        return True
    elif value.lower() == "false":
        return False
    raise Exception("Invalid boolean string passed as env var")


def main():
    start = time()
    registry = global_registry()
    config = pscr_config.Config.create(
        path.join(path.dirname(path.abspath(__file__)), "fetch_orgs_by_feature_config.ini")
    )
    _temp_setup_lambda()
    clear_logging_handlers()
    env_id = env.get("ENV_ID", "UNKNOWN")
    env_meta = MetadataProvider().get_metadata()
    svc_version = env_meta.image.rsplit(":", maxsplit=1)[-1]
    setup_logging(
        level=config_value(config, "LOG", "LEVEL"),
        svc_version=svc_version,
        env_id=env_id,
        svc_id="scheduler-fetch-orgs-by-features",
    )
    entitlement_service_auth_token = config_value(config, "ENTITLEMENT_SERVICE", "AUTH_TOKEN")
    if entitlement_service_auth_token == "use-k8-secrets":  # nosec
        api_key = env.get("ENTITLEMENT_SERVICE_AUTH_TOKEN_API_KEY")
        connector_id = env.get("ENTITLEMENT_SERVICE_AUTH_TOKEN_CONNECTOR_ID")
        entitlement_service_auth_token = f"{api_key}/{connector_id}"
    entitlement_service_base_url = config_value(config, "ENTITLEMENT_SERVICE", "BASE_URL")
    entitlement_service_results_limit = config_value(config, "ENTITLEMENT_SERVICE", "RESULTS_LIMIT")
    features = config_value(config, "ENTITLEMENT_SERVICE", "FEATURES").split(",")

    aws_s3_bucket = config_value(config, "AWS", "S3_BUCKET")
    if env.get("AWS_S3_BUCKET_LEGACY") not in ("use-config", None):
        aws_s3_bucket = env.get("AWS_S3_BUCKET_LEGACY")

    _logger.warning(
        f"Current config: {entitlement_service_base_url=} {entitlement_service_results_limit=} {features=} {aws_s3_bucket=}"
    )

    for feature in features:
        refresh_orgs_for_feature(
            feature,
            entitlement_service_base_url,
            entitlement_service_auth_token,
            entitlement_service_results_limit,
            aws_s3_bucket,
            registry,
        )

    if str_to_bool(env.get("ENTITLEMENT_SERVICE_GROUP_FEATURES", "true")):
        _logger.info("Uploading Org Feature Config file to S3")
        group_features_by_org(aws_s3_bucket)

    runtime = time() - start
    registry.gauge("duration").set_value(runtime)
    reporter = get_reporter(
        registry=registry,
        env=env_id,
        source=env_meta.task_id,
        server_type="scheduler.feature",
        version=svc_version,
        service_name_as_prefix=True,
    )
    if reporter:
        asyncio.get_event_loop().run_until_complete(reporter.report_now())


def lambda_handler(event, context):
    main()


if __name__ == "__main__":
    main()
